package POEx::ClockSkew;

use 5.008008;
use strict;
use warnings;

use POE;
use POE::Session::PlainCall;
use POE::Pipe::OneWay;
use Storable qw( dclone );
use POSIX::RT::Timer;
use POSIX;
use Time::HiRes qw( gettimeofday tv_interval );


our $VERSION = '0.0100';

#######################################
sub spawn
{
    my( $package, $args ) = @_;
    my $self = $package->new( $args );
    $self->init;
    POE::Session::PlainCall->create(
            object_states => [
                $self => [ qw( _start _stop shutdown skew piperead ) ]
            ] );
    return;
}

#######################################
sub debug
{
    my $self = shift;
    return unless $self->{debug};
    print STDERR "$self->{alias}: ", @_, "\n";
}

#######################################
sub new
{
    my( $self, $args ) = @_;
    return bless { %$args, args=>dclone $args }, $self;
}

#######################################
sub init
{
    my( $self ) = @_;
    $self->{format} = 'iFF';
    $self->{length} = length pack $self->{format}, 0, 0, 0;
    $self->{alias} ||= 'clock-skew';
    $self->{debug}  ||= 0;
    $self->{interval} ||= 5*60;        # 5 minutes;
    $self->{interval} = int $self->{interval};
    $self->{precision} ||= 1;
    $self->{signal} ||= SIGALRM;
}

#######################################
sub _start
{
    my( $self ) = @_;
    $self->debug( '_start' );
    $poe_kernel->alias_set( $self->{alias} );
    $poe_kernel->sig( shutdown => 'shutdown' );
    $self->build_pipe;
    $self->build_sig;
    $self->build_timer;
}

#######################################
sub _stop
{
    my( $self ) = @_;
    $self->debug( '_stop' );
}

#######################################
sub shutdown
{
    my( $self ) = @_;
    $self->{shutdown} = 1;
    $self->debug( 'shutdown' );
    $self->kill_timer;
    $self->kill_sig;
    $self->kill_pipe;
    $poe_kernel->alias_remove( $self->{alias} );
}

#######################################
sub build_pipe
{
    my( $self ) = @_;
    my( $read, $write ) = POE::Pipe::OneWay->new();
    die "Unable to build pipe: $!" unless defined $read;
    $self->{read} = $read;
    $self->{write} = $write;

    poe->kernel->select_read( $read, 'piperead' );
}

#######################################
sub kill_pipe
{
    my( $self ) = @_;
    poe->kernel->select_read( $self->{read} );
    delete $self->{read};
    delete $self->{write};
}

#######################################
sub piperead
{
    my( $self ) = @_;
    my( $epsilon, $expected ) = $self->read;
    $self->skew( $epsilon, $expected );
}

#######################################
sub skew
{
    my( $self, $epsilon, $expected ) = @_;
    $self->debug( join ' ', 'skew', $epsilon, $expected );
    $self->{detecting} = 0;
    $poe_kernel->signal( $poe_kernel, 'clock-skew', $epsilon, $expected );
}

##########################################################################
sub write
{
    my( $self, $one, $two ) = @_;
    return unless $self->{write};
    my $buffer = pack( $self->{format}, 0, $one, $two );
    syswrite( $self->{write}, $buffer, $self->{length} );
}

#######################################
sub read
{
    my( $self ) = @_;
    return unless $self->{read};
    my $buffer = '';
    sysread( $self->{read}, $buffer, $self->{length} );
    my( $type, $one, $two ) = unpack( $self->{format}, $buffer );
    $self->debug( "Read type=$type" ) unless $type == 0;
    return $one, $two;
}

##########################################################################
sub build_timer
{
    my( $self ) = @_;
    return if $self->{timer};
    $self->{timer} = POSIX::RT::Timer->new( 
                               value    => $self->{interval},
                               interval => $self->{interval},
                               signal   => $self->{signal},
                               clock    => 'monotonic'
                             );
    $self->now;
    return;    
}

sub kill_timer
{
    my( $self ) = @_;
    delete $self->{timer};
}

#######################################
sub build_sig
{
    my( $self ) = @_;
    return if $self->{sigact};

    use IO::File;
    my $fh;
    if( 1 ) {
        my $fh = IO::File->new;
        $fh->open( ">/tmp/clock-skew.log" );
        $fh->autoflush( 1 );
    }
    
    $self->{triggered} = 0;
    my $handler = sub { $self->signal_handler($fh); };
    $self->{flags} = 0;

    $self->{sigset} = POSIX::SigSet->new( $self->{signal} );
    $self->{sigact} = POSIX::SigAction->new( $handler, $self->{sigset}, $self->{flags} );
    $self->{sigact}->safe(1);
    sigaction( $self->{signal}, $self->{sigact} );
}

sub kill_sig
{
    my( $self ) = @_;
    return unless $self->{sigact};
    sigaction( $self->{signal}, undef() );
    delete $self->{sigact};
    delete $self->{sigset};
}


##########################################################################

#######################################
sub signal_handler
{
    my( $self, $fh ) = @_;
    $fh->print( "triggered\n" ) if $fh;
    $self->{debug} and warn "$$: Triggered\n";
    $self->detect_skew;
}

#######################################
sub detect_skew
{
    my( $self ) = @_;

    # XXX - check {timer}->get_overrun();

    my $ex = $self->calc_expected;
    $self->now;
    my $epsilon = $self->calc_epsilon( $ex );
    my $ep = abs( $epsilon );
    return if $ep < $self->{precision};

    my $expected = $ex->[0] + $ex->[1] / 1_000_000;
    $self->write( $epsilon, $expected );
}

#######################################
sub now
{
    my( $self ) = @_;
    $self->{now} = [ gettimeofday ];
}

#######################################
sub calc_expected
{
    my( $self ) = @_;
    return [ $self->{now}[0] + $self->{interval}, $self->{now}[1] ];
}

#######################################
sub calc_epsilon
{
    my( $self, $expected ) = @_;
    my $epsilon = tv_interval( $expected, [gettimeofday] );
    $self->{debug} and warn "$$: expected=$expected->[0] epsilon=$epsilon\n";
    return $epsilon;
}

1;

__END__

=head1 NAME

POEx::ClockSkew - Detect clock skew

=head1 SYNOPSIS

  use POEx::ClockSkew;

    POEx::ClockSkew->spawn({ interval => $seconds,
                         precision => 0.5,
                      });

    $poe_kernel->sig( 'clock-skew' => 'skewed' );


    sub skewed_handler
    {
        my( $self, $epsilon, $expected_time ) = @_[ OBJECT, ARG0, ARG1 ];
        # The system time has been skewed by $epsilon seconds
        # You might want to adjust your kernel timers
    }        


=head1 DESCRIPTION

POEx::ClockSkew implements a phase-locked loop that is used to detect jumps in
the system clock.  The phase-locked loop is a monotonic timer created with
L<POSIX::RT::Timer>.  This timer triggers a signal.  When the signal is
received, the current time is check against the expected time.  Any
discrepancy is due to system clock skew.

=head1 METHODS

=head2 spawn

=over 4

=item interval

=item precision

=item signal

=item alias

=back

=head1 SIGNALS

This component emits the following signals.

=head2 clock-skew

    $poe_kernel->sig( 'clock-skew' => 'skewed' );

    sub skewed_handler
    {
        my( $self, $epsilon, $expected_time ) = @_[ OBJECT, ARG0, ARG1 ];
    }

When 


=head2 shutdown

=head1 SEE ALSO

L<POE>, L<POSIX::RT::Timer>.

=head1 AUTHOR

Philip Gwyn, E<lt>gwyn-at-cpan.orgE<gt>

=head1 COPYRIGHT AND LICENSE

Copyright (C) 2012 by Philip Gwyn

This library is free software; you can redistribute it and/or modify
it under the same terms as Perl itself, either Perl version 5.8.8 or,
at your option, any later version of Perl 5 you may have available.


=cut
